#include "m_pd.h"
#include <julia/julia.h>
#include <unistd.h>

static t_class *julia_class;
static int initted;

typedef struct _julia
{
    t_object x_obj;
    t_canvas *x_canvas;
} t_julia;

static void *julia_new(void)
{
    t_julia *x = (t_julia *)pd_new(julia_class);
    outlet_new(&x->x_obj, &s_);
    x->x_canvas = canvas_getcurrent();
    if (!initted)
    {
        initted = 1;
    }
    return (x);
}

static void julia_eval(t_julia *x, t_symbol *s, int argc, t_atom *argv)
{
    char cmdbuf[1000];
    char atombuf[1000];
    int i;
    cmdbuf[0] = 0;
    for (i = 0; i < argc; i++)
    {
    	strcat(cmdbuf, " ");
    	atom_string(&argv[i], atombuf, 1000);
    	strcat(cmdbuf, atombuf);
    }
    strcat(cmdbuf, ";\n");
    jl_eval_string(cmdbuf);
}

static void julia_source(t_julia *x, t_symbol *filesym)
{
    char buf[MAXPDSTRING], *bufptr, filename[MAXPDSTRING], command[MAXPDSTRING];
    int fd = open_via_path(canvas_getdir(x->x_canvas)->s_name,
         filesym->s_name, "", buf, &bufptr, MAXPDSTRING, 0);
    if (fd < 0)
    {
        pd_error(x, "%s: can't open file", filesym->s_name);
        return;
    }
    close(fd);
    strncpy(filename, buf, MAXPDSTRING);
    filename[MAXPDSTRING-2] = 0;
    strcat(filename, "/");
    strncat(filename, bufptr, MAXPDSTRING-strlen(filename));
    filename[MAXPDSTRING-1] = 0;
    post("file: %s", filename);
    snprintf(command, MAXPDSTRING, "include(\"%s\");\n", filename);
    command[MAXPDSTRING-1] = 0;
    jl_eval_string(command);
}

static void julia_set(t_julia *x, t_symbol *s, int argc, t_atom *argv)
{
    char cmdbuf[MAXPDSTRING];
    int i;
    if (argc < 2 || argv[0].a_type != A_SYMBOL)
    {
        pd_error(s, "set: usage: array-name value...");
        return;
    }
    snprintf(cmdbuf, MAXPDSTRING, "%s = [", argv[0].a_w.w_symbol->s_name);
    cmdbuf[MAXPDSTRING-1] = 0;
    for (i = 1; i < argc; i++)
    {
    	atom_string(&argv[i], cmdbuf + strlen(cmdbuf),
            MAXPDSTRING-strlen(cmdbuf) - 5);
    	if (i < argc-1)
            strcat(cmdbuf, ",");
    }
    strcat(cmdbuf, "];\n");
    fprintf(stderr, "%s\n", cmdbuf);
    jl_eval_string(cmdbuf);
    if (jl_exception_occurred())
        post("exception occured in eval_string");
}

static void julia_get(t_julia *x, t_symbol *var, t_symbol *selector)
{
    if (jl_exception_occurred())
        post("exception occured before eval_string");
        
    jl_value_t *a = jl_eval_string(var->s_name);
    if (jl_exception_occurred())
        post("exception occured in eval_string");
    if (!jl_is_array(a))
    {
        post ("julia variable %s not found", var->s_name);
        return;
    }
    if (jl_array_eltype(a) != jl_float64_type && 
        jl_array_eltype(a) != jl_int64_type)
    {
        post("julia variable %s: bad array element type", var->s_name);
        post("type is %s", jl_typename_str(jl_array_eltype(a)));
        return;
    }
    {
        int n = jl_array_len(a), i, useint =
            (jl_array_eltype(a) == jl_int64_type);
        t_atom *at = alloca(n * sizeof(t_atom));
        for (i = 0; i < n; i++)
            SETFLOAT(at+i, 
                (useint ? ((long long *)jl_array_data(a))[i]
                    : ((double*)jl_array_data(a))[i]));
        outlet_list(x->x_obj.ob_outlet, 0, n, at);
    }
}

void julia_setup(void)
{
    julia_class = class_new(gensym("julia"), (t_newmethod)julia_new, 0,
        sizeof(t_julia), 0, A_NULL);
    class_addmethod(julia_class, (t_method)julia_eval,
        gensym("eval"), A_GIMME, A_NULL);
    class_addmethod(julia_class, (t_method)julia_source,
        gensym("source"), A_SYMBOL, A_NULL);
    class_addmethod(julia_class, (t_method)julia_get,
        gensym("get"), A_SYMBOL, A_DEFSYM, A_NULL);
    class_addmethod(julia_class, (t_method)julia_set,
        gensym("set"), A_GIMME, A_NULL);
    jl_init();
}
